var ErrorCodes = (function(){
    var errno = 0;
    return {
        NO_NICK: ++errno,
        NICK_TAKEN: ++errno,
        SYSTEM: ++errno,
        NO_TIMESTAMP: ++errno
    };
}());


function Chat(messages, input){
    this.messageWindow = messages;
    this.inputBox = input;
    this.runRequestXhr = null;
    this.runRequestSuccess = null;
    this.runRequestFailure = null;
    this.lastRequestArgs = null;
    this.didRetry = false;
    this.nick = '';
    this.nickGenerator = null;
    this.uid = null;
    this.lastPollTime = null;
    this.pollTimeoutId = null;
    this.pollCount = 0;
    this.audio = {};
    this.uidMap = {};
    this.needUidUpdate = true;
    this.prefix = './proxy/';

    input.addEventListener('keypress', this._inputKeypress.bind(this));
}

Chat.prototype = {
    _runRequestHandler: function(){
        if (this.runRequestXhr.readyState == 4){
            if (this.runRequestXhr.status == 200){
                var obj;
                try {
                    obj = JSON.parse(this.runRequestXhr.responseText);
                }
                catch (e){
                    obj = {
                        error: 'Unable to parse JSON',
                        code: ErrorCodes.INVALID_JSON_STRING
                    }
                }

                if (obj.error){
                    if (obj.code == ErrorCodes.UNKNOWN_USER){
                        if (!this.didRetry){
                            //Try and re-register
                            var runArgs = this.lastRequestArgs;
                            var oldNickGenerator = this.nickGenerator;
                            var tryCount = 0;
                            var oldNick = this.nick;
                            this.didRetry = true;

                            this.register(function(){
                                if (tryCount++ == 0){
                                    return oldNick;
                                }
                                else {
                                    return oldNickGenerator.call(this);
                                }
                            }, function(){
                                this.runRequest.apply(this, runArgs);
                            });
                        }
                        else {
                            this.didRetry = false;
                            if (this.runRequestFailure){
                                this.runRequestFailure.call(this, this.runRequestXhr, obj);
                            }
                        }
                    }
                    else if (this.runRequestFailure){
                        this.runRequestFailure.call(this, this.runRequestXhr, obj);
                    }
                }
                else if (!obj.error && this.runRequestSuccess){
                    this.didRetry = false;
                    this.runRequestSuccess.call(this, obj);
                }
            }
            else if (this.runRequestFailure){
                this.runRequestFailure.call(this, this.runRequestXhr, obj);
            }
        }
    },

    _inputKeypress: function(e){
        if (e.keyCode == 10 || e.keyCode == 13){
            e.preventDefault();

            var msg = this.inputBox.value;
            this.inputBox.value = '';

            msg = msg.replace(/^\s+$/, '');
            if (msg.length > 0){
                this.send(msg);
            }
        }
    },

    runRequest: function(method, url, obj, success, failure){
        if (!this.runRequestXhr){
            this.runRequestXhr = new XMLHttpRequest();
            this.runRequestXhr.onreadystatechange = this._runRequestHandler.bind(this);
            //this.runRequestXhr.responseType = 'text/json';
        }
        else if (this.runRequestXhr.readyState != 0 && this.runRequestXhr.readyState != 4){
            this.runRequestSuccess = null;
            this.runRequestFailure = null;
            this.runRequestXhr.abort();
        }

        this.runRequestSuccess = success;
        this.runRequestFailure = failure;
        this.runRequestXhr.open(method, this.prefix + url);
        this.runRequestXhr.setRequestHeader('X-Requested-With', 'XMLHttpRequest');

        if (obj){
            this.runRequestXhr.setRequestHeader('Content-type', 'text/json');
            obj = JSON.stringify(obj);
        }

        this.lastRequestArgs = arguments;
        this.runRequestXhr.send(obj);
    },

    isRegistered: function(){
        return this.uid !== null;
    },

    _registerSuccess: function(obj){
        this.nick = obj.nick;
        this.uid = obj.uid;
        this.uidMap[this.uid] = this.nick;

        if (obj.messages){
            for (var i = 0; i < obj.messages.length; i++){
                this.processMessage(obj.messages[i]);
            }
        }

        this.poll();
    },

    _registerFailure: function(xhr, obj){
        if (obj && (obj.code == ErrorCodes.NO_NICK || obj.code == ErrorCodes.NICK_TAKEN)){
            this.register(this.nickGenerator);
        }
        else {
            console.log('Registration failed');
            console.log(xhr, obj);
        }
    },

    register: function(nickGenerator){
        this.nickGenerator = nickGenerator;

        var nick = this.nickGenerator();
        if (nick !== undefined){
            this.runRequest('POST', '/register', {
                nick: nick
            }, this._registerSuccess.bind(this), this._registerFailure.bind(this));
        }
    },


    _zeroPad: function(n, len){
        var s = "" + n;
        while (s.length < len){
            s = "0" + s;
        }

        return s;
    },

    createTimeElement: function(ts){
        var time = document.createElement('time');
        var d = new Date(ts * 1000);

        var h = d.getHours();
        var m = d.getMinutes();
        var s = d.getSeconds();

        var datetimeAttr = d.getFullYear() + '-' + (d.getMonth() + 1) + '-' + d.getDate() + 'T' + h + ':' + m + ':' + s;
        var displayText = this._zeroPad(h, 2) + ':' + this._zeroPad(m, 2) + ':' + this._zeroPad(s, 2);

        time.setAttribute('datetime', datetimeAttr);
        time.appendChild(document.createTextNode(displayText));

        return time;
    },

    processMessage: function(obj){
        if (['say'].indexOf(obj.type) > -1){
            this.printMessage(obj);
        }
    },

    send: function(msg){
        var obj = {
            type: 'say',
            text: msg
        };

        this.runRequest('POST', '/messages?uid=' + encodeURIComponent(this.uid), obj, function(obj){
            for (var i = 0; i < obj.messages.length; i++){
                this.processMessage(obj.messages[i]);
            }
            this.poll();
        });
    },

    printMessage: function(obj){
        var div = document.createElement('div');
        div.className = 'message';

        var time = this.createTimeElement(obj.timestamp);
        var nick = document.createElement('span');
        nick.className = 'nick';
        nick.dataset.uid = obj.uid;

        if (obj.uid in this.uidMap){
            nick.appendChild(document.createTextNode(this.uidMap[obj.uid]));
        }
        else {
            this.needUidUpdate = true;
            nick.appendChild(document.createTextNode('Unknown'));
        }

        var msg = document.createElement('span');
        var html = this.htmlentities(obj.text);
        html = html.replace(/((https?|ftps?|ircs?):\/\/[^\s]*)/gi, function($0, $1){
            return '<a href="' + $1 + '" target="_blank">' + $1 + '</a>';
        });
        msg.innerHTML = html;

        div.appendChild(time);
        div.appendChild(nick);
        div.appendChild(msg);

        this.messageWindow.appendChild(div);
        this.messageWindow.scrollTop = this.messageWindow.scrollHeight;

        return div;
    },

    htmlentities: function(str){
        var charr = [];
        Array.prototype.forEach.call(str, function(ch){
            switch (ch){
                case '&':
                    ch = '&amp;';
                    break;
                case '<':
                    ch = '&lt;';
                    break;
                case '>':
                    ch = '&gt;';
                    break;
                case '"':
                    ch = '&quot;';
                    break;
            }

            charr.push(ch);
        });

        return charr.join('');
    },

    _pollSuccess: function(obj){
        for (var i = 0; i < obj.messages.length; i++){
            this.printMessage(obj.messages[i]);
        }

        if (i > 0){
            this.playAudio('message-received');
        }

        this.poll();
    },

    _pollFailure: function(xhr, obj){
        setTimeout(this.poll.bind(this), 2500);
    },

    poll: function(){
        if (this.needUidUpdate){
            this.runRequest('GET', '/userlist', null, function(obj){
                this.uidMap = obj.uidmap;
                this.needUidUpdate = false;
                try {
                    this.updateNicks();
                }
                catch (e){
                    console.log(e);
                }

                this.poll();
            });
        }
        else {
            this.runRequest('GET',
                '/messages?uid=' + encodeURIComponent(this.uid),
                null,
                this._pollSuccess.bind(this),
                this._pollFailure.bind(this));
        }
    },

    setAudio: function(event, audio){
        event = event.toLowerCase();
        event = event.replace(/[^a-z0-9]]/, '-');
        this.audio[event] = audio;
    },

    playAudio: function(event){
        if (this.audio[event]){
            this.audio[event].play();
        }
    },

    updateNicks: function(){
        var nicks = document.querySelectorAll('.nick');
        for (var i = 0; i < nicks.length; i++){
            while (nicks[i].firstChild) nicks[i].removeChild(nicks[i].firstChild);

            var uid = nicks[i].dataset.uid;
            if (uid in this.uidMap){
                nicks[i].appendChild(document.createTextNode(this.uidMap[uid]));
            }
            else {
                nicks[i].appendChild(document.createTextNode('Unknown'));
            }
        }
    }
};

window.addEventListener('load', function(){
    var nickLabel = document.getElementById('nick');
    var startUi = document.getElementById('startUi');
    var registerButton = document.getElementById('register');
    var chatUi = document.getElementById('chatUi');

    registerButton.addEventListener('click', function(){
        chatUi.style.display = 'block';
        startUi.style.display = 'none';

        var chatObj = new Chat(document.getElementById('messages'), document.getElementById('input'));
        chatObj.setAudio('message-received', document.getElementById('notify_sound'));
        chatObj.register(generateNick);
        document.getElementById('input').focus();

        function generateNick(){
            var num = Math.floor(100 + Math.random() * (999 - 100));
            var nick = prompt('Enter your nick: ', 'Guest' + num);
            if (!nick){
                return undefined;
            }
            else {
                nickLabel.textContent = nick;
                return nick;
            }
        }
    });
});
